@@ -167,7 +167,8 @@ predict.kernel.spline <- function(x,
                                   knots=c("quantiles","uniform"),
                                   basis=c("additive","tensor","glp"),
                                   model.return=FALSE,
-                                  tau=NULL){
+                                  tau=NULL,
+                                  weights=NULL){
 
   if(missing(x) || missing(y) || missing (K)) stop(" must provide x, y and K")
   if(!is.matrix(K)) stop(" K must be a two-column matrix")
@@ -191,7 +192,7 @@ predict.kernel.spline <- function(x,
 
   if(is.null(z)) {
 
-    ## First no categorical predictor case
+    ## First no categorical predictor case, never reached when called by crs()
 
     if(any(K[,1] > 0)) {
 
@@ -201,14 +202,14 @@ predict.kernel.spline <- function(x,
 
       if(basis=="additive" || basis=="glp") {
         if(is.null(tau))
-          model <- lm(y~P)
+          model <- lm(y~P,weights=weights)
         else
-          suppressWarnings(model <- rq(y~P,tau=tau,method="fn"))
+          suppressWarnings(model <- rq(y~P,tau=tau,method="fn",weights=weights))
       } else {
         if(is.null(tau))
-          model <- lm(y~P-1)
+          model <- lm(y~P-1,weights=weights)
         else
-          suppressWarnings(model <- rq(y~P-1,tau=tau,method="fn"))
+          suppressWarnings(model <- rq(y~P-1,tau=tau,method="fn",weights=weights))
       }
       if(is.null(xeval)) {
         fit.spline <- predict(model,interval="confidence",se.fit=TRUE)
@@ -222,9 +223,9 @@ predict.kernel.spline <- function(x,
       ## Degree == 0
 
       if(is.null(tau))
-        model <- lm(y~1)
+        model <- lm(y~1,weights=weights)
       else
-        suppressWarnings(model <- rq(y~1,tau=tau,method="fn"))
+        suppressWarnings(model <- rq(y~1,tau=tau,method="fn",weights=weights))
       if(is.null(xeval)) {
         fit.spline <- predict(model,interval="confidence",se.fit=TRUE)
       } else {
@@ -240,7 +241,7 @@ predict.kernel.spline <- function(x,
     if(is.null(tau))
       htt <- hatvalues(model)
     else
-      htt <- hat(model$x)
+      htt <- hat(model$qr)
 
     htt <- ifelse(htt < 1, htt, 1-.Machine$double.eps)
 
@@ -276,6 +277,7 @@ predict.kernel.spline <- function(x,
         for(i in 1:nrow.z.unique) {
           zz <- ind == ind.vals[i]
           L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
+          if(!is.null(weights)) L <- weights*L
           P <- prod.spline(x=x,K=K,knots=knots,basis=basis)
           k <- NCOL(P)
           if(basis=="additive" || basis=="glp") {
@@ -325,6 +327,7 @@ predict.kernel.spline <- function(x,
         for(i in 1:nrow.zeval.unique) {
           zz <- ind.zeval == ind.zeval.vals[i]
           L <- prod.kernel(Z=z,z=zeval.unique[ind.zeval.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
+          if(!is.null(weights)) L <- weights*L
           P <- prod.spline(x=x,K=K,knots=knots,basis=basis)
           k <- NCOL(P)
           if(basis=="additive" || basis=="glp") {
@@ -364,6 +367,7 @@ predict.kernel.spline <- function(x,
         for(i in 1:nrow.z.unique) {
           zz <- ind == ind.vals[i]
           L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
+          if(!is.null(weights)) L <- weights*L
           k <- 0
           ## Whether we use additive, glp, or tensor products, this
           ## model has no continuous predictors hence the intercept is
@@ -408,6 +412,7 @@ predict.kernel.spline <- function(x,
         for(i in 1:nrow.zeval.unique) {
           zz <- ind.zeval == ind.zeval.vals[i]
           L <- prod.kernel(Z=z,z=zeval.unique[ind.zeval.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
+          if(!is.null(weights)) L <- weights*L
           k <- 0
           if(is.null(tau))
             model.z.unique <- lm(y~x.intercept-1,weights=L)
@@ -472,7 +477,8 @@ deriv.kernel.spline <- function(x,
                                 basis=c("additive","tensor","glp"),
                                 deriv.index=1,
                                 deriv=0,
-                                tau=NULL) {
+                                tau=NULL,
+                                weights=NULL) {
 
   if(deriv == 0) stop(" deriv must be greater than zero")
 
@@ -493,7 +499,7 @@ deriv.kernel.spline <- function(x,
 
   if(basis=="additive" || basis=="glp") {
     K.additive <- K
-    K.additive[,2] <- ifelse(K[,1]==0,0,K[,2])    
+    K.additive[,2] <- ifelse(K[,1]==0,0,K[,2])
     K.additive[,1] <- ifelse(K[,1]>0,K[,1]-1,K[,1])
   }
 
@@ -512,9 +518,9 @@ deriv.kernel.spline <- function(x,
 
       if(basis=="additive") {
         if(is.null(tau))
-          model <- lm(y~P)
+          model <- lm(y~P,weights=weights)
         else
-          suppressWarnings(model <- rq(y~P,tau=tau,method="fn"))
+          suppressWarnings(model <- rq(y~P,tau=tau,method="fn",weights=weights))
 
         dim.P.deriv <- sum(K.additive[deriv.index,])
         deriv.start <- ifelse(deriv.index!=1,sum(K.additive[1:(deriv.index-1),])+1,1)
@@ -530,9 +536,9 @@ deriv.kernel.spline <- function(x,
         se.deriv <- sapply(1:NROW(P.deriv), function(i){ sqrt(P.deriv[i,deriv.ind.vec,drop=FALSE]%*%vcov.model[deriv.ind.vec,deriv.ind.vec]%*%t(P.deriv[i,deriv.ind.vec,drop=FALSE])) })
       } else if(basis=="tensor") {
         if(is.null(tau))
-          model <- lm(y~P-1)
+          model <- lm(y~P-1,weights=weights)
         else
-          suppressWarnings(model <- rq(y~P-1,tau=tau,method="fn"))
+          suppressWarnings(model <- rq(y~P-1,tau=tau,method="fn",weights=weights))
 
         deriv.spline <- P.deriv%*%coef(model)
 
@@ -544,9 +550,9 @@ deriv.kernel.spline <- function(x,
         se.deriv <- sapply(1:NROW(P.deriv), function(i){ sqrt(P.deriv[i,,drop=FALSE]%*%vcov.model%*%t(P.deriv[i,,drop=FALSE])) })
       } else if(basis=="glp") {
         if(is.null(tau))
-          model <- lm(y~P)
+          model <- lm(y~P,weights=weights)
         else
-          suppressWarnings(model <- rq(y~P,tau=tau,method="fn"))
+          suppressWarnings(model <- rq(y~P,tau=tau,method="fn",weights=weights))
         deriv.spline <- P.deriv%*%coef(model)[-1]
 
         if(is.null(tau))
@@ -588,6 +594,7 @@ deriv.kernel.spline <- function(x,
         for(i in 1:nrow.z.unique) {
           zz <- ind == ind.vals[i]
           L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
+          if(!is.null(weights)) L <- weights*L
           P <- prod.spline(x=x,K=K,knots=knots,basis=basis)
           P.deriv <- prod.spline(x=x,K=K,xeval=x[zz,,drop=FALSE],knots=knots,basis=basis,deriv.index=deriv.index,deriv=deriv)
           k <- NCOL(P)
@@ -596,9 +603,9 @@ deriv.kernel.spline <- function(x,
 
           if(basis=="additive") {
             if(is.null(tau))
-              model <- lm(y~P)
+              model <- lm(y~P,weights=L)
             else
-              suppressWarnings(model <- rq(y~P,tau=tau,method="fn"))
+              suppressWarnings(model <- rq(y~P,tau=tau,method="fn",weights=L))
             dim.P.deriv <- sum(K.additive[deriv.index,])
             deriv.start <- ifelse(deriv.index!=1,sum(K.additive[1:(deriv.index-1),])+1,1)
             deriv.end <- deriv.start+sum(K.additive[deriv.index,])-1
@@ -612,9 +619,9 @@ deriv.kernel.spline <- function(x,
             se.deriv[zz] <- sapply(1:NROW(P.deriv), function(i){ sqrt(P.deriv[i,deriv.ind.vec,drop=FALSE]%*%vcov.model[deriv.ind.vec,deriv.ind.vec]%*%t(P.deriv[i,deriv.ind.vec,drop=FALSE])) })
           } else if(basis=="tensor") {
             if(is.null(tau))
-              model <- lm(y~P-1)
+              model <- lm(y~P-1,weights=L)
             else
-              suppressWarnings(model <- rq(y~P-1,tau=tau,method="fn"))
+              suppressWarnings(model <- rq(y~P-1,tau=tau,method="fn",weights=L))
 
             deriv.spline[zz] <- P.deriv%*%coef(model)
 
@@ -626,9 +633,9 @@ deriv.kernel.spline <- function(x,
             se.deriv[zz] <- sapply(1:NROW(P.deriv), function(i){ sqrt(P.deriv[i,,drop=FALSE]%*%vcov.model%*%t(P.deriv[i,,drop=FALSE])) })
           } else if(basis=="glp") {
             if(is.null(tau))
-              model <- lm(y~P)
+              model <- lm(y~P,weights=L)
             else
-              suppressWarnings(model <- rq(y~P,tau=tau,method="fn"))
+              suppressWarnings(model <- rq(y~P,tau=tau,method="fn",weights=L))
 
             deriv.spline[zz] <- P.deriv%*%coef(model)[-1]
 
@@ -658,6 +665,7 @@ deriv.kernel.spline <- function(x,
         for(i in 1:nrow.zeval.unique) {
           zz <- ind.zeval == ind.zeval.vals[i]
           L <- prod.kernel(Z=z,z=zeval.unique[ind.zeval.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
+          if(!is.null(weights)) L <- weights*L
           P <- prod.spline(x=x,K=K,knots=knots,basis=basis)
           P.deriv <- prod.spline(x=x,K=K,xeval=xeval[zz,,drop=FALSE],knots=knots,basis=basis,deriv.index=deriv.index,deriv=deriv)
           k <- NCOL(P)
@@ -753,7 +761,8 @@ predict.factor.spline <- function(x,
                                   prune=FALSE,
                                   prune.index=NULL,
                                   trace=0,
-                                  tau=NULL){
+                                  tau=NULL,
+                                  weights=NULL){
 
   if(missing(x) || missing(y) || missing (K)) stop(" must provide x, y and K")
   if(!is.matrix(K)) stop(" K must be a two-column matrix")
@@ -787,33 +796,38 @@ predict.factor.spline <- function(x,
       ## a data frame so that we can readily determine columns that
       ## have been removed and assign logical values to all columns in
       ## P.
+
+      ## Note - this code is not reachable by crs() since pruning and
+      ## regression quantiles is not supported by stepCV (currently we
+      ## test and stop())
+
       P.df <- data.frame(P)
       names(P.df) <- paste("P",seq(1,NCOL(P.df)),sep="")
       if(basis=="additive" || basis=="glp") {
         if(is.null(tau))
-          model <- lm(y~.,data=P.df)
+          model <- lm(y~.,data=P.df,weights=weights)
         else
-          suppressWarnings(model <- rq(y~.,data=P.df,tau=tau,method="fn"))
+          suppressWarnings(model <- rq(y~.,data=P.df,tau=tau,method="fn",weights=weights))
       } else {
         if(is.null(tau))
-          model <- lm(y~.-1,data=P.df)
+          model <- lm(y~.-1,data=P.df,weights=weights)
         else
-          suppressWarnings(model <- rq(y~.-1,data=P.df,tau=tau,method="fn"))
+          suppressWarnings(model <- rq(y~.-1,data=P.df,tau=tau,method="fn",weights=weights))
       }
       if(is.null(tau))
         cv <- mean(residuals(model)^2/(1-hatvalues(model))^2)
       else
-        suppressWarnings(cv <- cv.rq(model,tau=tau))
+        suppressWarnings(cv <- cv.rq(model,tau=tau,weights=weights))
       console <- printClear(console)
       console <- printPush("Pruning...",console = console)
       if(basis=="additive" || basis=="glp") {
         if(is.null(tau))
-          model.pruned <- stepCV(lm(y~.,data=P.df),
+          model.pruned <- stepCV(lm(y~.,data=P.df,weights=weights),
                                  scope=list(upper=~.,lower=~1),
                                  k=log(length(y)),
                                  trace=trace)
         else
-          suppressWarnings(model.pruned <- stepCV(rq(y~.,data=P.df,tau=tau,method="fn"),
+          suppressWarnings(model.pruned <- stepCV(rq(y~.,data=P.df,tau=tau,method="fn",weights=weights),
                                                   scope=list(upper=~.,lower=~1),
                                                   k=log(length(y)),
                                                   trace=trace))
@@ -821,12 +835,12 @@ predict.factor.spline <- function(x,
 
       } else {
         if(is.null(tau))
-          model.pruned <- stepCV(lm(y~.-1,data=P.df),
+          model.pruned <- stepCV(lm(y~.-1,data=P.df,weights=weights),
                                  scope=list(upper=~.,lower=~1),
                                  k=log(length(y)),
                                  trace=trace)
         else
-          suppressWarnings(model.pruned <- stepCV(rq(y~.-1,data=P.df,tau=tau,method="fn"),
+          suppressWarnings(model.pruned <- stepCV(rq(y~.-1,data=P.df,tau=tau,method="fn",weights=weights),
                                                   scope=list(upper=~.,lower=~1),
                                                   k=log(length(y)),
                                                   trace=trace))
@@ -835,35 +849,35 @@ predict.factor.spline <- function(x,
       if(is.null(tau))
         cv.pruned <- mean(residuals(model.pruned)^2/(1-hatvalues(model.pruned))^2)
       else
-        suppressWarnings(cv.pruned <- cv.rq(model.pruned,tau=tau))
+        suppressWarnings(cv.pruned <- cv.rq(model.pruned,tau=tau,weights=weights))
 
       if(cv.pruned <= cv) {
         IND <- logical()
         for(i in 1:NCOL(P.df)) IND[i] <- any(names(P.df)[i]==names(model.pruned$model[,-1,drop=FALSE]))
         if(basis=="additive" || basis=="glp") {
           if(is.null(tau))
-            model <- lm(y~P[,IND,drop=FALSE])
+            model <- lm(y~P[,IND,drop=FALSE],weights=weights)
           else
-            suppressWarnings(model <- rq(y~P[,IND,drop=FALSE],tau=tau,method="fn"))
+            suppressWarnings(model <- rq(y~P[,IND,drop=FALSE],tau=tau,method="fn",weights=weights))
         } else {
           if(is.null(tau))
-            model <- lm(y~P[,IND,drop=FALSE]-1)
+            model <- lm(y~P[,IND,drop=FALSE]-1,weights=weights)
           else
-            suppressWarnings(model <- rq(y~P[,IND,drop=FALSE]-1,tau=tau,method="fn"))
+            suppressWarnings(model <- rq(y~P[,IND,drop=FALSE]-1,tau=tau,method="fn",weights=weights))
         }
       } else {
         warning(" pruned model did not lower cross-validation score, using non-pruned bases")
         IND <- !logical(length=NCOL(P))
         if(basis=="additive" || basis=="glp") {
           if(is.null(tau))
-            model <- lm(y~P)
+            model <- lm(y~P,weights=weights)
           else
-            suppressWarnings(model <- rq(y~P,tau=tau,method="fn"))
+            suppressWarnings(model <- rq(y~P,tau=tau,method="fn",weights=weights))
         } else {
           if(is.null(tau))
-            model <- lm(y~P-1)
+            model <- lm(y~P-1,weights=weights)
           else
-            suppressWarnings(model <- rq(y~P-1,tau=tau,method="fn"))
+            suppressWarnings(model <- rq(y~P-1,tau=tau,method="fn",weights=weights))
         }
       }
     } else if(prune) {
@@ -871,39 +885,49 @@ predict.factor.spline <- function(x,
       IND <- prune.index
       if(basis=="additive" || basis=="glp") {
         if(is.null(tau))
-          model <- lm(y~P[,IND,drop=FALSE])
+          model <- lm(y~P[,IND,drop=FALSE],weights=weights)
         else
-          suppressWarnings(model <- rq(y~P[,IND,drop=FALSE],tau=tau,method="fn"))
+          suppressWarnings(model <- rq(y~P[,IND,drop=FALSE],tau=tau,method="fn",weights=weights))
       } else {
         if(is.null(tau))
-          model <- lm(y~P[,IND,drop=FALSE]-1)
+          model <- lm(y~P[,IND,drop=FALSE]-1,weights=weights)
         else
-          suppressWarnings(model <- rq(y~P[,IND,drop=FALSE]-1,tau=tau,method="fn"))
+          suppressWarnings(model <- rq(y~P[,IND,drop=FALSE]-1,tau=tau,method="fn",weights=weights))
       }
       cv <- NULL
       if(is.null(tau))
         cv.pruned <- mean(residuals(model)^2/(1-hatvalues(model))^2)
       else
-        suppressWarnings(cv.pruned <- cv.rq(model,tau=tau))
+        suppressWarnings(cv.pruned <- cv.rq(model,tau=tau,weights=weights))
     } else {
-      ## No pruning
+      ## No pruning, default case
       IND <- !logical(length=NCOL(P))
       if(basis=="additive" || basis=="glp") {
         if(is.null(tau))
-          model <- lm(y~P)
+          model <- lm(y~P,weights=weights)
         else
-          suppressWarnings(model <- rq(y~P,tau=tau,method="fn"))
+          suppressWarnings(model <- rq(y~P,tau=tau,method="fn",weights=weights))
       } else {
         if(is.null(tau))
-          model <- lm(y~P-1)
+          model <- lm(y~P-1,weights=weights)
         else
-          suppressWarnings(model <- rq(y~P-1,tau=tau,method="fn"))
+          suppressWarnings(model <- rq(y~P-1,tau=tau,method="fn",weights=weights,x=TRUE))
       }
       cv.pruned <- NULL
-      if(is.null(tau))
+      if(is.null(tau)) {
         cv <- mean(residuals(model)^2/(1-hatvalues(model))^2)
-      else
-        suppressWarnings(cv <- cv.rq(model,tau=tau))
+      } else {
+        if(basis=="additive" || basis=="glp")
+          model.hat <- lm(y~P,weights=weights)
+        else
+          model.hat <- lm(y~P-1,weights=weights)
+        htt <- hat(model.hat$qr)
+        ## Note - this is defined in util.R so if you modify there you must modify here also
+        if(is.null(weights))
+          cv <- mean(check.function(residuals(model),tau)/(1-htt)^(1/sqrt(tau*(1-tau))))
+        else
+          cv <- mean(check.function(residuals(model)*sqrt(weights),tau)/(1-htt)^(1/sqrt(tau*(1-tau))))
+      }
     }
 
     if(is.null(xeval)) {
@@ -918,16 +942,19 @@ predict.factor.spline <- function(x,
     ## Degree == 0, no pruning possible
     IND <- TRUE
 
-    if(is.null(tau))
-      model <- lm(y~1)
-    else
-      suppressWarnings(model <- rq(y~1,tau=tau,method="fn"))
-
-
-    if(is.null(tau))
+    if(is.null(tau)) {
+      model <- lm(y~1,weights=weights)
       cv <- mean(residuals(model)^2/(1-hatvalues(model))^2) ## Added
-    else
-      suppressWarnings(cv <- cv.rq(model,tau=tau))
+    } else {
+      suppressWarnings(model <- rq(y~1,tau=tau,method="fn",weights=weights))
+      model.hat <- lm(y~1,weights=weights)
+      htt <- hat(model.hat$qr)
+      ## Note - this is defined in util.R so if you modify there you must modify here also
+      if(is.null(weights))
+        cv <- mean(check.function(residuals(model),tau)/(1-htt)^(1/sqrt(tau*(1-tau))))
+      else
+        cv <- mean(check.function(residuals(model)*sqrt(weights),tau)/(1-htt)^(1/sqrt(tau*(1-tau))))
+    }
 
     cv.pruned <- NULL
     if(is.null(xeval)) {
@@ -949,7 +976,7 @@ predict.factor.spline <- function(x,
   if(is.null(tau))
     htt <- hatvalues(model)
   else
-    htt <- hat(model$x)
+    htt <- hatvalues(model.hat)
 
   return(list(fitted.values=fit.spline,
               df.residual=model$df.residual,
@@ -983,7 +1010,8 @@ deriv.factor.spline <- function(x,
                                 deriv.index=1,
                                 deriv=0,
                                 prune.index=NULL,
-                                tau=NULL) {
+                                tau=NULL,
+                                weights=NULL) {
 
   if(missing(x) || missing(y) || missing (K)) stop(" must provide x, y and K")
   if(deriv == 0) stop(" derivative must be a positive integer")
@@ -1002,7 +1030,7 @@ deriv.factor.spline <- function(x,
 
   if(basis=="additive" || basis=="glp") {
     K.additive <- K
-    K.additive[,2] <- ifelse(K[,1]==0,0,K[,2])    
+    K.additive[,2] <- ifelse(K[,1]==0,0,K[,2])
     K.additive[,1] <- ifelse(K[,1]>0,K[,1]-1,K[,1])
   }
   if(K[deriv.index,1]!=0) {
@@ -1026,9 +1054,9 @@ deriv.factor.spline <- function(x,
     vcov.mat.model <- matrix(0,nrow=NCOL(P),ncol=NCOL(P))
     if(basis=="additive") {
       if(is.null(tau))
-        model <- lm(y~P[,prune.index,drop=FALSE])
+        model <- lm(y~P[,prune.index,drop=FALSE],weights=weights)
       else
-        suppressWarnings(model <- rq(y~P[,prune.index,drop=FALSE],tau=tau,method="fn"))
+        suppressWarnings(model <- rq(y~P[,prune.index,drop=FALSE],tau=tau,weights=weights,method="fn"))
 
       coef.vec.model[prune.index] <- coef(model)[-1]
 
@@ -1044,9 +1072,9 @@ deriv.factor.spline <- function(x,
       deriv.ind.vec <- ifelse(prune.index,deriv.ind.vec,FALSE)
     } else if(basis=="tensor") {
       if(is.null(tau))
-        model <- lm(y~P[,prune.index,drop=FALSE]-1)
+        model <- lm(y~P[,prune.index,drop=FALSE]-1,weights=weights)
       else
-        suppressWarnings(model <- rq(y~P[,prune.index,drop=FALSE]-1,tau=tau,method="fn"))
+        suppressWarnings(model <- rq(y~P[,prune.index,drop=FALSE]-1,tau=tau,weights=weights,method="fn"))
       coef.vec.model[prune.index] <- coef(model)
 
       if(is.null(tau))
@@ -1058,9 +1086,9 @@ deriv.factor.spline <- function(x,
       deriv.ind.vec <- ifelse(prune.index,deriv.ind.vec,FALSE)
     } else if(basis=="glp") {
       if(is.null(tau))
-        model <- lm(y~P[,prune.index,drop=FALSE])
+        model <- lm(y~P[,prune.index,drop=FALSE],weights=weights)
       else
-        suppressWarnings(model <- rq(y~P[,prune.index,drop=FALSE],tau=tau,method="fn"))
+        suppressWarnings(model <- rq(y~P[,prune.index,drop=FALSE],tau=tau,weights=weights,method="fn"))
       coef.vec.model[prune.index] <- coef(model)[-1]
 
       if(is.null(tau))
@@ -1108,7 +1136,10 @@ cv.kernel.spline.wrapper <- function(x,
                                      knots=c("quantiles","uniform","auto"),
                                      basis=c("additive","tensor","glp"),
                                      cv.func=c("cv.ls","cv.gcv","cv.aic"),
-                                     tau=NULL) {
+                                     cv.df.min=1,
+                                     tau=NULL,
+                                     weights=NULL,
+                                     singular.ok=FALSE) {
 
   knots.opt <- knots;
 
@@ -1129,7 +1160,10 @@ cv.kernel.spline.wrapper <- function(x,
                            knots="quantiles",
                            basis=basis,
                            cv.func=cv.func,
-                           tau=tau)
+                           cv.df.min=cv.df.min,
+                           tau=tau,
+                           weights=weights,
+                           singular.ok=singular.ok)
 
     cv.uniform <- cv.kernel.spline(x=x,
                                    y=y,
@@ -1144,7 +1178,10 @@ cv.kernel.spline.wrapper <- function(x,
                                    knots="uniform",
                                    basis=basis,
                                    cv.func=cv.func,
-                                   tau=tau)
+                                   cv.df.min=cv.df.min,
+                                   tau=tau,
+                                   weights=weights,
+                                   singular.ok=singular.ok)
     if(cv > cv.uniform) {
       cv <- cv.uniform
       knots.opt <- "uniform"
@@ -1165,7 +1202,10 @@ cv.kernel.spline.wrapper <- function(x,
                            knots=knots,
                            basis=basis,
                            cv.func=cv.func,
-                           tau=tau)
+                           cv.df.min=cv.df.min,
+                           tau=tau,
+                           weights=weights,
+                           singular.ok=singular.ok)
 
   }
 
@@ -1198,7 +1238,10 @@ cv.kernel.spline <- function(x,
 														 knots=c("quantiles","uniform"),
 														 basis=c("additive","tensor","glp"),
 														 cv.func=c("cv.ls","cv.gcv","cv.aic"),
-                             tau=NULL) {
+                             cv.df.min=1,
+                             tau=NULL,
+                             weights=NULL,
+                             singular.ok=FALSE) {
 
   if(missing(x) || missing(y) || missing (K)) stop(" must provide x, y and K")
 
@@ -1216,13 +1259,16 @@ cv.kernel.spline <- function(x,
 
   ## Check dimension of P prior to calculating the basis
 
-  if(dim.bs(basis=basis,kernel=TRUE,degree=K[,1],segments=K[,2]) >= (n-1))
+  if(n - dim.bs(basis=basis,kernel=TRUE,degree=K[,1],segments=K[,2]) <= cv.df.min)
     return(sqrt(.Machine$double.xmax))
 
   ## Otherwise, compute the cross-validation function
 
   if(is.null(z)) {
-    ## No categorical predictors
+    ## Here we need to use lm.wfit throughout with weights, but lm.wfit
+    ## needs non-null weights, so if weights are null create a vector of
+    ## ones
+    ## No categorical predictors, never reached when called by crs()
     if(any(K[,1] > 0)) {
       ## Check for rank-deficient fit
       P <- prod.spline(x=x,K=K,knots=knots,basis=basis)
@@ -1230,32 +1276,44 @@ cv.kernel.spline <- function(x,
       if(NCOL(P) >= (n-1))
         return(sqrt(.Machine$double.xmax))
       if(basis=="additive" || basis=="glp") {
+        ## Test for full column rank
+        if(!singular.ok) {
+          if(!is.fullrank(cbind(1,P)))
+            return(sqrt(.Machine$double.xmax))
+        }
         ## Additive spline regression models have an intercept in the lm()
         ## model (though not in the gsl.bs function)
         if(is.null(tau)) {
-          epsilon <- residuals(model <- lm.fit(cbind(1,P),y))
-          if(model$rank < (NCOL(P)+1))
-            return(sqrt(.Machine$double.xmax))
+          if(is.null(weights))
+            epsilon <- residuals(model <- lm.fit(cbind(1,P),y))
+          else
+            epsilon <- residuals(model <- lm.wfit(cbind(1,P),y,weights))
         } else {
-          ## Test for full column rank
-          if(!is.fullrank(cbind(1,P)))
-            return(sqrt(.Machine$double.xmax))         
-          residuals <- tryCatch(residuals(rq.fit(cbind(1,P),y,tau=tau,method="fn")),error=function(e){FALSE})
+          if(is.null(weights))
+            residuals <- tryCatch(residuals(rq.fit(cbind(1,P),y,tau=tau,method="fn")),error=function(e){FALSE})
+          else
+            residuals <- tryCatch(residuals(rq.wfit(cbind(1,P),y,tau=tau,weights,method="fn")),error=function(e){FALSE})
           if(is.logical(residuals))
-            return(sqrt(.Machine$double.xmax))            
+            return(sqrt(.Machine$double.xmax))
         }
       } else {
-        if(is.null(tau)) {
-          epsilon <- residuals(model <- lm.fit(P,y))
-          if(model$rank < NCOL(P))
-            return(sqrt(.Machine$double.xmax))
-        } else {
-          ## Test for full column rank
+        ## Test for full column rank
+        if(!singular.ok) {
           if(!is.fullrank(P))
-            return(sqrt(.Machine$double.xmax))         
-          residuals <- tryCatch(residuals(rq.fit(P,y,tau=tau,method="fn")),error=function(e){FALSE})
+            return(sqrt(.Machine$double.xmax))
+        }
+        if(is.null(tau)) {
+          if(is.null(weights))
+            epsilon <- residuals(model <- lm.fit(P,y))
+          else
+            epsilon <- residuals(model <- lm.wfit(P,y,weights))
+        } else {
+          if(is.null(weights))
+            residuals <- tryCatch(residuals(rq.fit(P,y,tau=tau,method="fn")),error=function(e){FALSE})
+          else
+            residuals <- tryCatch(residuals(rq.wfit(P,y,tau=tau,weights,method="fn")),error=function(e){FALSE})
           if(is.logical(residuals))
-            return(sqrt(.Machine$double.xmax))            
+            return(sqrt(.Machine$double.xmax))
         }
       }
       htt <- hat(P)
@@ -1281,34 +1339,35 @@ cv.kernel.spline <- function(x,
        for(i in 1:nrow.z.unique) {
         zz <- ind == ind.vals[i]
         L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
+        if(!is.null(weights)) L <- weights*L
         if(basis=="additive" || basis=="glp") {
+          ## Test for full column rank
+          if(!singular.ok) {
+            if(!is.fullrank(cbind(1,P)*L))
+              return(sqrt(.Machine$double.xmax))
+          }
           ## Additive spline regression models have an intercept in
           ## the lm() model (though not in the gsl.bs function)
           if(is.null(tau)) {
             model <- lm.wfit(cbind(1,P),y,L)
-            if(model$rank < (NCOL(P)+1))
-              return(sqrt(.Machine$double.xmax))
           } else {
-            ## Test for full column rank
-            if(!is.fullrank(cbind(1,P)*L))
-              return(sqrt(.Machine$double.xmax))         
             model <- tryCatch(rq.wfit(cbind(1,P),y,weights=L,tau=tau,method="fn"),error=function(e){FALSE})
             if(is.logical(model))
-              return(sqrt(.Machine$double.xmax))            
+              return(sqrt(.Machine$double.xmax))
             model.hat <- lm.wfit(cbind(1,P),y,L)
           }
         } else {
+          ## Test for full column rank
+          if(!singular.ok) {
+            if(!is.fullrank(P*L))
+              return(sqrt(.Machine$double.xmax))
+          }
           if(is.null(tau)) {
             model <- lm.wfit(P,y,L)
-            if(model$rank < NCOL(P))
-              return(sqrt(.Machine$double.xmax))
           } else {
-            ## Test for full column rank
-            if(!is.fullrank(P*L))
-              return(sqrt(.Machine$double.xmax))         
             model <- tryCatch(rq.wfit(P,y,weights=L,tau=tau,method="fn"),error=function(e){FALSE})
             if(is.logical(model))
-              return(sqrt(.Machine$double.xmax))            
+              return(sqrt(.Machine$double.xmax))
             model.hat <- lm.wfit(P,y,L)
           }
         }
@@ -1318,7 +1377,7 @@ cv.kernel.spline <- function(x,
         else
           htt[zz] <- hat(model.hat$qr)[zz]
       }
-      
+
     } else {
       ## No predictors for which degree > 0
       z.factor <- data.frame(factor(z[,1]))
@@ -1326,9 +1385,12 @@ cv.kernel.spline <- function(x,
       for(i in 1:nrow.z.unique) {
         zz <- ind == ind.vals[i]
         L <- prod.kernel(Z=z,z=z.unique[ind.vals[i],],lambda=lambda,is.ordered.z=is.ordered.z)
+        if(!is.null(weights)) L <- weights*L
         ## Test for full column rank
-        if(!is.fullrank(matrix(1,n,1)*L))
-          return(sqrt(.Machine$double.xmax))                        
+        if(!singular.ok) {
+          if(!is.fullrank(matrix(1,n,1)*L))
+            return(sqrt(.Machine$double.xmax))
+        }
         ## Whether we use additive, glp, or tensor products, this
         ## model has no continuous predictors hence the intercept is
         ## the parameter that may shift with the categorical
@@ -1339,7 +1401,7 @@ cv.kernel.spline <- function(x,
         } else {
           model <- tryCatch(rq.wfit(matrix(1,n,1),y,weights=L,tau=tau,method="fn"),error=function(e){FALSE})
           if(is.logical(model))
-            return(sqrt(.Machine$double.xmax))            
+            return(sqrt(.Machine$double.xmax))
           model.hat <- lm.wfit(matrix(1,n,1),y,L)
           htt[zz] <- hat(model.hat$qr)[zz]
         }
@@ -1351,18 +1413,22 @@ cv.kernel.spline <- function(x,
 
   }
 
+  ## If weights exist, need to use weighted residuals
+
+  if(!is.null(weights)) epsilon <- epsilon*sqrt(weights)
+
   if(cv.func == "cv.ls") {
-    if(is.null(tau)) 
+    if(is.null(tau))
       cv <- mean(epsilon^2/(1-htt)^2)
     else
       ## Note - this is defined in util.R so if you modify there you must modify here also
       cv <- mean(check.function(epsilon,tau)/(1-htt)^(1/sqrt(tau*(1-tau))))
   } else if(cv.func == "cv.gcv"){
-    if(is.null(tau)) 
+    if(is.null(tau))
       cv <- mean(epsilon^2/(1-mean(htt))^2)
     else
-      ## Note - this is defined in util.R so if you modify there you must modify here also        
-      cv <- mean(check.function(epsilon,tau)/(1-mean(htt))^(1/sqrt(tau*(1-tau))))        
+      ## Note - this is defined in util.R so if you modify there you must modify here also
+      cv <- mean(check.function(epsilon,tau)/(1-mean(htt))^(1/sqrt(tau*(1-tau))))
   } else if(cv.func == "cv.aic"){
     traceH <- sum(htt)
     if(is.null(tau)) {
@@ -1374,7 +1440,7 @@ cv.kernel.spline <- function(x,
     }
     cv <- ifelse(penalty < 0, .Machine$double.xmax, log(sigmasq)+penalty);
   }
-  
+
   return(ifelse(!is.na(cv),cv,.Machine$double.xmax))
 
 }
@@ -1392,7 +1458,10 @@ cv.factor.spline.wrapper <- function(x,
 																		 knots=c("quantiles","uniform","auto"),
 																		 basis=c("additive","tensor","glp"),
 																		 cv.func=c("cv.ls","cv.gcv","cv.aic"),
-                                     tau=NULL) {
+                                     cv.df.min=1,
+                                     tau=NULL,
+                                     weights=NULL,
+                                     singular.ok=FALSE) {
 
   knots.opt <- knots
 
@@ -1408,7 +1477,10 @@ cv.factor.spline.wrapper <- function(x,
                            knots="quantiles",
                            basis=basis,
                            cv.func=cv.func,
-                           tau=tau)
+                           cv.df.min=cv.df.min,
+                           tau=tau,
+                           weights=weights,
+                           singular.ok=singular.ok)
 
     cv.uniform <- cv.factor.spline(x=x,
                                    y=y,
@@ -1418,7 +1490,10 @@ cv.factor.spline.wrapper <- function(x,
                                    knots="uniform",
                                    basis=basis,
                                    cv.func=cv.func,
-                                   tau=tau)
+                                   cv.df.min=cv.df.min,
+                                   tau=tau,
+                                   weights=weights,
+                                   singular.ok=singular.ok)
     if(cv > cv.uniform) {
       cv <- cv.uniform
       knots.opt <- "uniform"
@@ -1434,7 +1509,10 @@ cv.factor.spline.wrapper <- function(x,
                            knots=knots,
                            basis=basis,
                            cv.func=cv.func,
-                           tau=tau)
+                           cv.df.min=cv.df.min,
+                           tau=tau,
+                           weights=weights,
+                           singular.ok=singular.ok)
 
   }
 
@@ -1456,7 +1534,10 @@ cv.factor.spline <- function(x,
 														 knots=c("quantiles","uniform"),
 														 basis=c("additive","tensor","glp"),
 														 cv.func=c("cv.ls","cv.gcv","cv.aic"),
-                             tau=NULL) {
+                             cv.df.min=1,
+                             tau=NULL,
+                             weights=NULL,
+                             singular.ok=FALSE) {
 
   if(missing(x) || missing(y) || missing (K)) stop(" must provide x, y and K")
   if(!is.matrix(K)) stop(" K must be a two-column matrix")
@@ -1477,7 +1558,7 @@ cv.factor.spline <- function(x,
     for(i in NCOL(z)) categories[i] <- length(unique(z[,i]))
   }
 
-  if(dim.bs(basis=basis,kernel=TRUE,degree=K[,1],segments=K[,2],include=I,categories=categories) >= (n-1))
+  if(n - dim.bs(basis=basis,kernel=TRUE,degree=K[,1],segments=K[,2],include=I,categories=categories) <= cv.df.min)
     return(sqrt(.Machine$double.xmax))
 
   ## Otherwise, compute the cross-validation function
@@ -1490,34 +1571,52 @@ cv.factor.spline <- function(x,
     if(NCOL(P) >= (n-1))
       return(sqrt(.Machine$double.xmax))
     if(basis=="additive" || basis=="glp") {
+      ## Test for full column rank
+      if(!singular.ok) {
+        if(!is.fullrank(cbind(1,P)))
+          return(sqrt(.Machine$double.xmax))
+      }
       ## Additive spline regression models have an intercept in the
       ## lm() model (though not in the gsl.bs function)
       if(is.null(tau)) {
-        model <- lm.fit(cbind(1,P),y)
-        ## Test for rank-deficient fit
-        if(model$rank < (NCOL(P)+1)) return(sqrt(.Machine$double.xmax))
+        if(is.null(weights))
+          model <- lm.fit(cbind(1,P),y)
+        else
+          model <- lm.wfit(cbind(1,P),y,weights)
       } else {
-        ## Test for full column rank
-        if(!is.fullrank(cbind(1,P)))
-          return(sqrt(.Machine$double.xmax))      
-        model <- tryCatch(rq.fit(cbind(1,P),y,tau=tau,method="fn"),error=function(e){FALSE})
+        if(is.null(weights))
+          model <- tryCatch(rq.fit(cbind(1,P),y,tau=tau,method="fn"),error=function(e){FALSE})
+        else
+          model <- tryCatch(rq.wfit(cbind(1,P),y,tau=tau,weights,method="fn"),error=function(e){FALSE})
         if(is.logical(model))
-          return(sqrt(.Machine$double.xmax))            
-        model.hat <- lm.fit(cbind(1,P),y)
+          return(sqrt(.Machine$double.xmax))
+        if(is.null(weights))
+          model.hat <- lm.fit(cbind(1,P),y)
+        else
+          model.hat <- lm.wfit(cbind(1,P),y,weights)
       }
     } else {
-      if(is.null(tau)) {
-        model <- lm.fit(P,y)
-        ## Test for rank-deficient fit
-        if(model$rank < NCOL(P)) return(sqrt(.Machine$double.xmax))
-      } else {
-        ## Test for full column rank
+      ## Test for full column rank
+      if(!singular.ok) {
         if(!is.fullrank(P))
-          return(sqrt(.Machine$double.xmax))      
-        model <- tryCatch(rq.fit(P,y,tau=tau,method="fn"),error=function(e){FALSE})
+          return(sqrt(.Machine$double.xmax))
+      }
+      if(is.null(tau)) {
+        if(is.null(weights))
+          model <- lm.fit(P,y)
+        else
+          model <- lm.wfit(P,y,weights)
+      } else {
+        if(is.null(weights))
+          model <- tryCatch(rq.fit(P,y,tau=tau,method="fn"),error=function(e){FALSE})
+        else
+          model <- tryCatch(rq.wfit(P,y,tau=tau,weights,method="fn"),error=function(e){FALSE})
         if(is.logical(model))
-          return(sqrt(.Machine$double.xmax))            
-        model.hat <- lm.fit(P,y)
+          return(sqrt(.Machine$double.xmax))
+        if(is.null(weights))
+          model.hat <- lm.fit(P,y)
+        else
+          model.hat <- lm.wfit(P,y,weights)
       }
     }
     if(is.null(tau))
@@ -1531,18 +1630,22 @@ cv.factor.spline <- function(x,
     epsilon <- y-mean(y)
   }
 
+  ## If weights exist, need to use weighted residuals
+
+  if(!is.null(weights)) epsilon <- epsilon*sqrt(weights)
+
   if(cv.func == "cv.ls") {
-    if(is.null(tau)) 
+    if(is.null(tau))
       cv <- mean(epsilon^2/(1-htt)^2)
     else
       ## Note - this is defined in util.R so if you modify there you must modify here also
       cv <- mean(check.function(epsilon,tau)/(1-htt)^(1/sqrt(tau*(1-tau))))
   } else if(cv.func == "cv.gcv"){
-    if(is.null(tau)) 
+    if(is.null(tau))
       cv <- mean(epsilon^2/(1-mean(htt))^2)
     else
-      ## Note - this is defined in util.R so if you modify there you must modify here also        
-      cv <- mean(check.function(epsilon,tau)/(1-mean(htt))^(1/sqrt(tau*(1-tau))))        
+      ## Note - this is defined in util.R so if you modify there you must modify here also
+      cv <- mean(check.function(epsilon,tau)/(1-mean(htt))^(1/sqrt(tau*(1-tau))))
   } else if(cv.func == "cv.aic"){
     traceH <- sum(htt)
     if(is.null(tau)) {
@@ -1554,7 +1657,7 @@ cv.factor.spline <- function(x,
     }
     cv <- ifelse(penalty < 0, .Machine$double.xmax, log(sigmasq)+penalty);
   }
-  
+
   return(ifelse(!is.na(cv),cv,.Machine$double.xmax))
 
 }
