@@ -46,11 +46,11 @@
 #include "Cache_File_Point.hpp"
 #include "Set_Element.hpp"
 
-#ifdef WINDOWS 
-#ifndef isnan
-inline bool isnan ( double x ) { return x != x; }
-#endif
-#endif
+/*#ifdef WINDOWS */ 
+/*  #ifndef isnan */ //we change the name "isnan" to nomad_isnan  to avoid redefined.
+inline bool nomad_isnan ( double x ) { return x != x; }
+/*  #endif 
+#endif */
 
 namespace NOMAD {
 
